using System;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Bespoke.Games.Framework
{
    /// <summary>
    /// 
    /// </summary>
    public class Actor
    {
        #region Events

        /// <summary>
        /// 
        /// </summary>
        public event CustomDrawHandler CustomDraw;

        /// <summary>
        /// 
        /// </summary>
        public event CustomUpdateHandler CustomUpdate;

        #endregion

        #region Properties

        /// <summary>
        /// 
        /// </summary>
        public string Name
        {
            get
            {
                return mName;
            }
            set
            {
                mName = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public bool Visible
        {
            get
            {
                return mVisible;
            }
            set
            {
                mVisible = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Vector3 Position
        {
            get
            {
                return mPosition;
            }
            set
            {
                mPosition = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Vector3 Orientation
        {
            get
            {
                return mOrientation;
            }
            set
            {
                mOrientation = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public float MovementVelocity
        {
            get
            {
                return mMovementVelocity;
            }
            set
            {
                mMovementVelocity = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public float Scale
        {
            get
            {
                return mScale;
            }
            set
            {
                mScale = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Model Model
        {
            get
            {
                return mModel;
            }
            set
            {
                mModel = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public CameraComponent CameraComponent
        {
            get
            {
                return mCameraComponent;
            }
            set
            {
                mCameraComponent = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Matrix WorldMatrix
        {
            get
            {
                return mWorldMatrix;
            }
            set
            {
                mWorldMatrix = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Matrix ProjectionMatrix
        {
            get
            {
                return mProjectionMatrix;
            }
            set
            {
                mProjectionMatrix = value;
            }
        }

        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="name"></param>
        /// <param name="position"></param>
        /// <param name="orientation"></param>
        /// <param name="scale"></param>
        /// <param name="model"></param>
        /// <param name="camera"></param>
        /// <param name="projectionMatrix"></param>
        public Actor(string name, Vector3 position, Vector3 orientation, float scale,
            Model model, CameraComponent camera, Matrix projectionMatrix)
        {
            mName = name;
            mVisible = true;
            mPosition = position;
            mOrientation = orientation;
            mScale = scale;
            mModel = model;
            mCameraComponent = camera;
            mProjectionMatrix = projectionMatrix;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="mesh"></param>
        /// <returns></returns>
        public static CustomModelData GetCustomModelData(ModelMesh mesh)
        {
            return mesh.Tag as CustomModelData;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T">The type of vertices to return.</typeparam>
        /// <param name="graphicsDevice"></param>
        /// <param name="mesh"></param>
        /// <returns></returns>
        public static T[] GetMeshVertices<T>(GraphicsDevice graphicsDevice, CustomModelData customModelData, int vertexDataClassSize) where T : struct
        {
            if (customModelData == null)
            {
                throw new ArgumentNullException("customModelData");
            }

            byte[] data = customModelData.VertexData;
            T[] meshVertices = new T[data.Length / vertexDataClassSize];

            using (VertexBuffer vertexBuffer = new VertexBuffer(graphicsDevice, data.Length, ResourceUsage.None))
            {
                vertexBuffer.SetData<byte>(data, 0, data.Length, SetDataOptions.None);                
                vertexBuffer.GetData<T>(meshVertices);
            }

            return meshVertices;
        }

        /// <summary>
        /// 
        /// </summary>
        public virtual void Initialize()
        {
            mBoneTransforms = new Matrix[mModel.Bones.Count];
            mModel.CopyAbsoluteBoneTransformsTo(mBoneTransforms);

            foreach (ModelMesh mesh in mModel.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                }
            }
        }

        /// <summary>
        /// Allows the actor to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public virtual void Update(GameTime gameTime)
        {
            if (CustomUpdate != null)
            {
                CustomUpdate(this, new CustomUpdateEventArgs(gameTime));
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="gameTime"></param>
        public virtual void Draw(GameTime gameTime)
        {
            if (CustomDraw != null)
            {
                CustomDraw(this, new CustomDrawEventArgs(gameTime));
            }
            else
            {
                if (mVisible)
                {
                    foreach (ModelMesh mesh in mModel.Meshes)
                    {
                        foreach (BasicEffect effect in mesh.Effects)
                        {
                            effect.World = mBoneTransforms[mesh.ParentBone.Index] * Matrix.CreateFromAxisAngle(Vector3.UnitX, mOrientation.X) * Matrix.CreateFromAxisAngle(Vector3.UnitY, mOrientation.Y) * Matrix.CreateFromAxisAngle(Vector3.UnitZ, mOrientation.Z) * Matrix.CreateScale(mScale) * Matrix.CreateTranslation(mPosition);
                            effect.View = mCameraComponent.ViewMatrix;
                            effect.Projection = mProjectionMatrix;
                        }

                        mesh.Draw();
                    }
                }
            }
        }

        protected string mName;
        protected bool mVisible;
        protected Vector3 mPosition;
        protected Vector3 mOrientation;
        protected float mMovementVelocity;
        protected float mScale;
        protected Model mModel;
        protected Matrix[] mBoneTransforms;

        protected CameraComponent mCameraComponent;
        protected Matrix mWorldMatrix;
        protected Matrix mProjectionMatrix;
    }

    /// <summary>
    /// 
    /// </summary>
    public class ActorList : List<Actor> { }
}