using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Bespoke.Games.Framework
{
    /// <summary>
    /// This is a game component that implements IUpdateable.
    /// </summary>
    public class CameraComponent : Microsoft.Xna.Framework.GameComponent
    {
        #region Properties

        /// <summary>
        /// 
        /// </summary>
        public KeyboardComponent KeyboardComponent
        {
            get
            {
                return mKeyboardComponent;
            }
            set
            {
                mKeyboardComponent = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public MouseComponent MouseComponent
        {
            get
            {
                return mMouseComponent;
            }
            set
            {
                mMouseComponent = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public GamePadComponent GamePadComponent
        {
            get
            {
                return mGamePadComponent;
            }
            set
            {
                mGamePadComponent = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Actor AttachedActor
        {
            get
            {
                return mAttachedActor;
            }
            set
            {
                mAttachedActor = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Vector3 AttachedActorOffset
        {
            get
            {
                return mAttachedActorOffset;
            }
            set
            {
                mAttachedActorOffset = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Vector3 Position
        {
            get
            {
                return mPosition;
            }
            set
            {
                mPosition = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Vector3 Orientation
        {
            get
            {
                return mOrientation;
            }
            set
            {
                mOrientation = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public float MovementVelocity
        {
            get
            {
                return mMovementVelocity;
            }
            set
            {
                mMovementVelocity = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public float RotationVelocity
        {
            get
            {
                return mRotationVelocity;
            }
            set
            {
                mRotationVelocity = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Matrix ViewMatrix
        {
            get
            {
                return mViewMatrix;
            }
            set
            {
                mViewMatrix = value;
            }
        }

        #endregion

        public CameraComponent(Game game)
            : base(game)
        {
            mMovementVelocity = 1.0f;
            mRotationVelocity = MathHelper.Pi / 180.0f;
        }

        /// <summary>
        /// Allows the game component to perform any initialization it needs to before starting
        /// to run.  This is where it can query for any required services and load content.
        /// </summary>
        public override void Initialize()
        {            
            mViewMatrix = Matrix.CreateTranslation(mPosition);

            base.Initialize();
        }

        /// <summary>
        /// Allows the game component to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Update(GameTime gameTime)
        {
            Vector3 position;
            Vector3 orientation;

            if (mAttachedActor != null)
            {
                position = mAttachedActor.Position + mAttachedActorOffset;
                orientation = mAttachedActor.Orientation;
            }
            else
            {
                Move();
                Rotate();
                position = mPosition;
                orientation = mOrientation;
            }

            mViewMatrix = Matrix.CreateRotationX(orientation.X) * Matrix.CreateRotationY(orientation.Y) * Matrix.CreateRotationZ(orientation.Z) * Matrix.CreateTranslation(position);

            base.Update(gameTime);
        }

        /// <summary>
        /// 
        /// </summary>
        public void Reset()
        {
            mPosition = Vector3.Zero;
            mOrientation = Vector3.Zero;
        }

        #region Private Methods

        /// <summary>
        /// 
        /// </summary>
        private void Move()
        {
            Vector3 delta = Vector3.Zero;

            if (mKeyboardComponent != null)
            {
                if (mKeyboardComponent.IsKeyDown(Keys.W))
                {
                    delta.Z = mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.S))
                {
                    delta.Z = -mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.A))
                {
                    delta.X = mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.D))
                {
                    delta.X = -mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.PageUp))
                {
                    delta.Y = -mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.PageDown))
                {
                    delta.Y = mMovementVelocity;
                }

                if (mKeyboardComponent.IsKeyDown(Keys.X))
                {
                    Reset();
                }
            }

            if (mMouseComponent != null)
            {
                if (mMouseComponent.RelativeScrollWheelValue < 0)
                {
                    delta.Y = -mMovementVelocity;
                }

                if (mMouseComponent.RelativeScrollWheelValue > 0)
                {
                    delta.Y = mMovementVelocity;
                }
            }

            mPosition += delta;
        }

        /// <summary>
        /// 
        /// </summary>
        private void Rotate()
        {
            Vector3 delta = Vector3.Zero;

            if ((mMouseComponent != null) && (mMouseComponent.IsLeftButtonHeldDown))
            {
                Vector2 mouseVector = new Vector2(-(float)mMouseComponent.AbsolutePositionX / (float)Game.Window.ClientBounds.Width* 2f + 1f, (float)mMouseComponent.AbsolutePositionY / (float)Game.Window.ClientBounds.Height * 2f - 1f);
                if (mouseVector.X < 0)
                {
                    delta.Y = -mouseVector.X * RotationVelocity;
                }
                else
                {
                    delta.Y = -mouseVector.X * RotationVelocity;
                }

                if (mouseVector.Y < 0)
                {
                    delta.X = mouseVector.Y * mRotationVelocity;
                }
                else
                {
                    delta.X = mouseVector.Y * mRotationVelocity;
                }
            }

            mOrientation += delta;
            if (mOrientation.Y > sMaxOrientation.Y)
            {
                mOrientation.Y = sMaxOrientation.Y;
            }
            if (mOrientation.Y < sMinOrientation.Y)
            {
                mOrientation.Y = sMinOrientation.Y;
            }

            if (mOrientation.X > sMaxOrientation.X)
            {
                mOrientation.X = sMaxOrientation.X;
            }
            if (mOrientation.X < sMinOrientation.X)
            {
                mOrientation.X = sMinOrientation.X;
            }

            if (mOrientation.Z > sMaxOrientation.Z)
            {
                mOrientation.Z = sMaxOrientation.Z;
            }
            if (mOrientation.Z < sMinOrientation.Z)
            {
                mOrientation.Z = sMinOrientation.Z;
            }
        }

        #endregion

        private static readonly Vector3 sMaxOrientation = new Vector3(1.5f, 1.5f, 0f);
        private static readonly Vector3 sMinOrientation = new Vector3(-1.5f, -1.5f, 0f);

        private KeyboardComponent mKeyboardComponent;
        private MouseComponent mMouseComponent;
        private GamePadComponent mGamePadComponent;

        private Vector3 mPosition;
        private Vector3 mOrientation;
        private float mMovementVelocity;
        private float mRotationVelocity;
        private Matrix mViewMatrix;

        private Actor mAttachedActor;
        private Vector3 mAttachedActorOffset;
    }
}