using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace Bespoke.Games.Framework
{
    /// <summary>
    ///
    /// </summary>
    public class GridComponent : Microsoft.Xna.Framework.DrawableGameComponent
    {
        #region Properties

        /// <summary>
        /// 
        /// </summary>
        public CameraComponent CameraComponent
        {
            get
            {
                return mCameraComponent;
            }
            set
            {
                mCameraComponent = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Color Color
        {
            get
            {
                return mColor;
            }
            set
            {
                if (mColor != value)
                {
                    mColor = value;
                    InitGrid();
                }
            }
        }

        /// <summary>
        /// Determines the number of grid cells e.g. 16 = 16x16 grid
        /// </summary>
        public int Size
        {
            get
            {
                return mSize;
            }
            set
            {
                if (mScale != value)
                {
                    mSize = value;
                    InitGrid();
                }
            }
        }

        /// <summary>
        /// Determines the spacing of each line e.g. 32 means 32 units between each line.
        /// </summary>
        public int Scale
        {
            get
            {
                return mScale;
            }
            set
            {
                if (mScale != value)
                {
                    mScale = value;
                    InitGrid();
                }
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Matrix WorldMatrix
        {
            get
            {
                return mWorldMatrix;
            }
            set
            {
                mWorldMatrix = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Matrix ProjectionMatrix
        {
            get
            {
                return mProjectionMatrix;
            }
            set
            {
                mProjectionMatrix = value;
            }
        }

        #endregion

        public GridComponent(Game game)
            : base(game)
        {
            mColor = Color.Wheat;
            mSize = 16;
            mScale = 32;            
        }

        /// <summary>
        /// Allows the game component to perform any initialization it needs to before starting
        /// to run.  This is where it can query for any required services and load content.
        /// </summary>
        public override void Initialize()
        {
            IGraphicsDeviceService graphicsService = (IGraphicsDeviceService)Game.Services.GetService(typeof(IGraphicsDeviceService));
            mGraphicsDevice = graphicsService.GraphicsDevice;
            mContentManager = new ContentManager(Game.Services);

            mVertexDeclaration = new VertexDeclaration(mGraphicsDevice, VertexPositionColor.VertexElements);
            mPositionColorEffect = mContentManager.Load<Effect>(@"Content\Effects\PositionColor");            
            InitGrid();

            base.Initialize();
        }

        /// <summary>
        /// Allows the game component to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Update(GameTime gameTime)
        {
            if (Visible)
            {
                mPositionColorEffect.Parameters["WorldViewProjection"].SetValue(mWorldMatrix * mCameraComponent.ViewMatrix * mProjectionMatrix);
            }

            base.Update(gameTime);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Draw(GameTime gameTime)
        {
            mGraphicsDevice.VertexDeclaration = mVertexDeclaration;
            mGraphicsDevice.RenderState.CullMode = CullMode.None;

            mPositionColorEffect.Begin();
            foreach (EffectPass effectPass in mPositionColorEffect.CurrentTechnique.Passes)
            {
                effectPass.Begin();

                mGraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, mGridPoints, 0, (mSize + 1) * 2);

                effectPass.End();
            }

            mPositionColorEffect.End();

            base.Draw(gameTime);
        }

        /// <summary>
        /// 
        /// </summary>
        private void InitGrid()
        {
            mGridPoints = new VertexPositionColor[4 * (mSize + 1)];

            float adjustedScale = mScale * 0.1f;
            float maxPosition = mSize * adjustedScale / 2;
            for (int i = 0, j = 0; i < mSize + 1; i++, j = 4 * i)
            {
                float position = maxPosition - (i * adjustedScale);

                // Vertical line
                mGridPoints[j] = new VertexPositionColor(new Vector3(position, 0, maxPosition), mColor);
                mGridPoints[j + 1] = new VertexPositionColor(new Vector3(position, 0, -maxPosition), mColor);

                // Horizontal line
                mGridPoints[j + 2] = new VertexPositionColor(new Vector3(maxPosition, 0, position), mColor);
                mGridPoints[j + 3] = new VertexPositionColor(new Vector3(-maxPosition, 0, position), mColor);
            }
        }

        private VertexDeclaration mVertexDeclaration;        
        private GraphicsDevice mGraphicsDevice;
        private ContentManager mContentManager;

        private CameraComponent mCameraComponent;
        private Effect mPositionColorEffect;
        private Matrix mWorldMatrix;
        private Matrix mProjectionMatrix;

        private VertexPositionColor[] mGridPoints;
        private Color mColor;
        private int mSize;
        private int mScale;
    }
}


