using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Bespoke.Games.Framework
{
    /// <summary>
    /// 
    /// </summary>
    public static class ScreenCapture
    {
        /// <summary>
        /// 
        /// </summary>
        public static string BaseFileName
        {
            get
            {
                return sBaseFileName;
            }
            set
            {
                sBaseFileName = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public static ImageFormat ImageFormat
        {
            get
            {
                return sImageFormat;
            }
            set
            {
                sImageFormat = value;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        static ScreenCapture()
        {
            sBaseFileName = "ScreenCapture";
            sImageFormat = ImageFormat.Bmp;
        }

        /// <summary>
        /// 
        /// </summary>
        public static void CaptureToFile()
        {
            WindowCapture.CaptureToFile(CaptureFileName, sImageFormat);
        }

        /// <summary>
        /// 
        /// </summary>
        private static string CaptureFileName
        {
            get
            {
                return String.Format("{0}{1}.{2}", sBaseFileName, sCount++, sImageFormat.ToString().ToLower());
            }
        }

        private static string sBaseFileName;
        private static int sCount;
        private static ImageFormat sImageFormat;
    }

    /// <summary>
    /// Functions for capturing the content of a window and saving it to an image file.
    /// </summary>
    internal static class WindowCapture
    {
        /// <summary>
        /// Captures the content of the currently active window and saves it to an image
        /// file of the specified type.
        /// </summary>
        /// <param name="fileName">The name of the file to create. The file extension must be specified.</param>
        /// <param name="imageFormat">The format of the file to create.</param>
        public static void CaptureToFile(string fileName, ImageFormat imageFormat)
        {
            CaptureToFile(fileName, imageFormat, User32.GetActiveWindow());
        }

        /// <summary>
        /// Captures the contents of the specified window and saves it
        /// to an image file with the specified name and format.
        /// </summary>
        /// <param name="fileName">The name of the file to create. The appropriate extension must be specified.</param>
        /// <param name="imageFormat">The format of the file to create.</param>
        public static void CaptureToFile(string fileName, ImageFormat imageFormat, IntPtr hWnd)
        {
            if (hWnd == IntPtr.Zero)
            {
                hWnd = User32.GetActiveWindow();
            }
        
            RECT windowRect;
            RECT clientRect;
            User32.GetWindowRect(hWnd, out windowRect);
            User32.GetClientRect(hWnd, out clientRect);

            int borderSizeX = 0;
            int borderSizeY = 0;
            uint WS_BORDER = 0x00800000;
            if ((User32.GetWindowStyle(hWnd) & WS_BORDER) != 0)
            {
                int SM_CXDLGFRAME = 7;
                int SM_CYDLGFRAME = 8;
                borderSizeX = User32.GetSystemMetrics(SM_CXDLGFRAME);
                borderSizeY = User32.GetSystemMetrics(SM_CYDLGFRAME);
            }

            IntPtr hdcSrc = User32.GetWindowDC(hWnd);
            IntPtr hdcDest = GDI32.CreateCompatibleDC(hdcSrc);
            IntPtr hDestBitmap = GDI32.CreateCompatibleBitmap(hdcSrc, clientRect.Width, clientRect.Height);
            GDI32.SelectObject(hdcDest, hDestBitmap);

            int SRCCOPY = 0x00CC0020;
            GDI32.BitBlt
                (
                hdcDest,
                0,
                0,
                clientRect.Width, clientRect.Height,
                hdcSrc,
                windowRect.Width - clientRect.Width - borderSizeX,
                windowRect.Height - clientRect.Height - borderSizeY,
                SRCCOPY
                );

            SaveImage(hDestBitmap, fileName, imageFormat);

            User32.ReleaseDC(hWnd, hdcSrc);
            GDI32.DeleteDC(hdcDest);
            GDI32.DeleteObject(hDestBitmap);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hBitmap"></param>
        /// <param name="fileName"></param>
        /// <param name="imageFormat"></param>
        private static void SaveImage(IntPtr hBitmap, string fileName, ImageFormat imageFormat)
        {
            Bitmap fromHBitmap = Image.FromHbitmap(hBitmap);
            Bitmap newBitmap = new Bitmap(fromHBitmap, fromHBitmap.Width, fromHBitmap.Height);
            newBitmap.Save(fileName, imageFormat);
            newBitmap.Dispose();
        }
    }

    /// <summary>
    /// 
    /// </summary>
    internal class GDI32
    {
        [DllImport("GDI32.dll")]
        public static extern bool BitBlt
            (
            IntPtr hdcDest,
            int xDest, int yDest,
            int width, int height,
            IntPtr hdcSrc,
            int xSrc, int ySrc,
            int dwRop
            );

        [DllImport("GDI32.dll")]
        public static extern IntPtr CreateCompatibleBitmap(IntPtr hdc, int width, int height);

        [DllImport("GDI32.dll")]
        public static extern IntPtr CreateCompatibleDC(IntPtr hdc);

        [DllImport("GDI32.dll")]
        public static extern bool DeleteDC(IntPtr hdc);

        [DllImport("GDI32.dll")]
        public static extern bool DeleteObject(IntPtr hObject);

        [DllImport("GDI32.dll")]
        public static extern int GetDeviceCaps(IntPtr hdc, int index);

        [DllImport("GDI32.dll")]
        public static extern IntPtr SelectObject(IntPtr hdc, IntPtr hgdiobj);
    }

    /// <summary>
    /// 
    /// </summary>
    [StructLayout(LayoutKind.Sequential)]
    internal struct RECT
    {
        public int left;
        public int top;
        public int right;
        public int bottom;

        /// <summary>
        /// 
        /// </summary>
        public int Width
        {
            get
            {
                return this.right - this.left; 
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public int Height
        {
            get
            {
                return this.bottom - this.top;
            }
        }
    }

    /// <summary>
    /// 
    /// </summary>
    internal class User32
    {
        [DllImport("User32.dll")]
        public static extern IntPtr GetDesktopWindow();

        [DllImport("User32.dll")]
        public static extern IntPtr GetActiveWindow();

        [DllImport("User32.dll")]
        public static extern IntPtr GetWindowDC(IntPtr hWnd);

        [DllImport("User32.dll")]
        public static extern IntPtr ReleaseDC(IntPtr hWnd, IntPtr hdc);

        [DllImport("User32.dll")]
        public static extern bool GetWindowRect(IntPtr hWnd, out RECT rect);

        [DllImport("User32.dll")]
        public static extern bool GetClientRect(IntPtr hWnd, out RECT rect);

        [DllImport("User32.dll")]
        public static extern int GetSystemMetrics(int index);

        public static uint GetWindowStyle(IntPtr hWnd)
        {
            int GWL_STYLE = -16;
            return GetWindowLong(hWnd, GWL_STYLE);
        }

        [DllImport("User32.dll")]
        public static extern uint GetWindowLong(IntPtr hWnd, int index);
    }
}
